#ifndef __LIST_H
#define __LIST_H

#include <iostream>
using namespace std;

/*** Declarations: ***/

template <class content_t>
struct cons_cell {
    content_t car;
    cons_cell<content_t>* cdr;
    cons_cell(content_t , cons_cell<content_t>* = 0);
};

template <class content_t>
class list {
    cons_cell<content_t>* clone_cons (cons_cell<content_t>*) const;
    
 protected:
    
    cons_cell<content_t>* content;
  
 public:
    list(void);
    list(const list&);
    list(cons_cell<content_t>*);
    ~list(void);
    const list& operator=(const list&);

    int null(void) const;
    content_t car(void) const;
    void cdr(void);
    void cons(content_t);
    void rmth(int = 0);
    void print(ostream&) const;
    
    // Iterator on lists:
    class iterator {
        cons_cell<content_t>* place;

    public:
        iterator(cons_cell<content_t> *);
        const iterator& operator=(cons_cell<content_t>*);

        iterator(const iterator&);
        const iterator& operator=(const iterator&);

        void operator++(void);
        content_t& operator*(void);
        int operator==(const iterator&) const;
        int operator!=(const iterator&) const;
    };

    // start and end iterators:
    iterator begin(void);
    iterator end(void);
};

template <class content_t>
ostream& operator<< (ostream&, const list<content_t>&);

/*** Definitions: ***/

template <class content_t>
cons_cell<content_t>::cons_cell (content_t val, cons_cell<content_t>* rest) {
    car = val;
    cdr = rest;
}

template <class content_t>
list<content_t>::list (void) {
    content = 0;
}

template <class content_t>
list<content_t>::list (cons_cell<content_t>* c)
    : content (c) {
    /* empty */
}

template <class content_t>
list<content_t>::list(const list<content_t>& l) {
    content = clone_cons(l.content);
}

/*
 * Does the deep copying of content.  We cannot easily do it with a
 * cycle, since a naive such a cycle will copy the list in the wrong
 * order.  So we write a recursive function.
 */
template <class content_t>
cons_cell<content_t>* list<content_t>::clone_cons (cons_cell<content_t>* c) const {
    if (c == 0) return 0;
    return (new cons_cell<content_t>(c -> car, clone_cons(c -> cdr)));
}

template <class content_t>
const list<content_t>& list<content_t>::operator=(const list<content_t>& rhs) {
    if (this != &rhs) // Standard alias test (when we do a = a;)
        // Delete old content
        while (content != 0)
            cdr();
        content = clone_cons(rhs.content);
    return *this;  // because we may need to do a = b = c;
}

template <class content_t>
list<content_t>::~list (void) {
    while (content != 0)
        cdr();
}

template <class content_t>
int list<content_t>::null (void) const {
    return content == 0;
}

template <class content_t>
content_t list<content_t>::car (void) const {
    return content -> car;
}

template <class content_t>
void list<content_t>::cdr (void) {
    if (content != 0) {
        cons_cell<content_t>* tmp = content;
        content = content -> cdr;
        delete tmp;
    }
}

template <class content_t>
void list<content_t>::cons(content_t c) {
    content = new cons_cell<content_t>(c,content);
}

template <class content_t>
void list<content_t>::rmth (int which) {
    cons_cell<content_t>* place = content;
    for (int i = 0; i < which - 1; i++) { // go to element which - 1...
        if (place == 0)
            return;  // nothing to delete, we are done.
        place = place -> cdr;
    }
    if (place !=0 && place -> cdr != 0) {
        cons_cell<content_t>* to_delete = place -> cdr;
        place -> cdr = place -> cdr -> cdr;
        delete to_delete;
    }
}

template <class content_t>
void list<content_t>::print(ostream& out) const {
    cons_cell<content_t>* iter = content;
    out << "(";
    while (iter != 0) {
        out << iter -> car;
        iter = iter -> cdr;
        if (iter != 0)
            out << ",";
    }
    out << ")";
}

template <class content_t>
ostream& operator<< (ostream& out, const list<content_t>& value) {
    value.print(out);
    return out;
}

// Iterators:

template <class content_t>
list<content_t>::iterator::iterator(cons_cell<content_t> * p) { 
    place = p; 
}

template <class content_t>
list<content_t>::iterator::iterator(const list<content_t>::iterator& rhs) { 
    place = rhs.place; 
}

template <class content_t>
const typename list<content_t>::iterator&
list<content_t>::iterator::operator=(cons_cell<content_t> * p) { 
    place = p;
    return *this;
}

template <class content_t>
const typename list<content_t>::iterator&
list<content_t>::iterator::operator=(const list<content_t>::iterator& rhs) { 
    place = rhs.place;
    return *this;
}

template <class content_t>
void list<content_t>::iterator::operator++(void) { 
    place = place -> cdr;
}

template <class content_t>
content_t& list<content_t>::iterator::operator*(void) { 
    return place -> car; 
}

template <class content_t>
int list<content_t>::iterator::operator==(const list<content_t>::iterator& rhs) const {
    return place == rhs.place; 
}

template <class content_t>
int list<content_t>::iterator::operator!=(const list<content_t>::iterator& rhs) const {
    return !operator==(rhs); 
}

template <class content_t>
typename list<content_t>::iterator  list<content_t>::begin(void) {
    typename list<content_t>::iterator ret(content);
    return ret;
}

template <class content_t>
typename list<content_t>::iterator list<content_t>::end(void) {
    typename list<content_t>::iterator ret(0);
    return ret;
}

#endif /* __LIST_H */
